import greenfoot.*;

/**
 * SUPPORT CLASS: HexWorld (World subclass)<br>
 * AUTHOR NAME: danpost (Greenfoot.org username)<br>
 * VERSION DATE: January 11, 2014<br><br>
 * 
 * DESCRIPTION: Creates a hexagonal world (a world filled with hexagons)<br>
 */
public class HexWorld extends World
{
    private int hexSize = 40; // the minimum length of the side of a square that can wrap each hexagon
    private int hexWidth, hexHeight;
    private Color hexBackgroundColor = Color.WHITE, // background color of active hexagons
                  hexFrameColor = Color.GRAY, // frame color of active hexagons
                  inactiveFillColor = Color.GRAY; // fill color of inactive hexagons
    private boolean rotated; // flag indicating whether hexagons are laying down or not

    /**
     * HexWorld Constructor: creates a world of sufficient size, then saves the necessary state values and calls a method to
     * create the background image of the world.
     *
     * @param across the number of columns of hexagons to draw
     * @param down the number of rows of hexagons to draw
     * @param hexLength the length from one point of a hexagon to the opposite point (vertically or horizontally) of that hexagon;<br>
     * NOTE: the length must be an even number; when divisible by four, outer rows are longer, and when not, outer rows are shorter
     * @param toBeRotated a flag to indicate rotation of hexagon (true means the sides are to be vertical and the hexagons will stand on points)
     */
    public HexWorld(int across, int down, int hexLength, boolean toBeRotated)
    {
        // create a world of sufficient size for desired hexagonal world
        super(across*hexLength*3/(toBeRotated ? 3 : 4)+hexLength*2+(toBeRotated ? 0 : hexLength/4),
            down*hexLength*3/(toBeRotated ? 4 : 3)+hexLength*2+(toBeRotated ? hexLength/4 : 0),
            1);
        // fail if the size of the hexagon is invalid (must be at least 4 and even)
        if (hexLength < 4 || hexLength%2 == 1)
        {
            String text = "The size of the hexagons must be a multiple of two.";
            text += "\n\nThe outer row will be longer when a multiple of four;";
            text += "\nthe outer row will be shorter when not a multiple of four.\n\n\n";
            // System.out.println(text);
            Greenfoot.stop();
            return;
        }
        // save state values
        hexWidth = across;
        hexHeight = down;
        hexSize = hexLength;
        rotated = toBeRotated;
        // create image of the hexagonal world on the world canvas
        paintBackground();
    }

    /**
     * Used to paint (or re-paint) the current hexagonal world on the world canvas
     */
    public void paintBackground()
    {
        // fill background with inactive fill color
        getBackground().setColor(inactiveFillColor);
        getBackground().fill();
        // create the hexagon to draw on world canvas
        GreenfootImage image = getHexImage();
        // draw specified grid of hexagons on world canvas
        for (int i=0; i<hexWidth; i++) for (int j=0; j<hexHeight; j++)
        {
            if (!isHexLocation(i, j)) continue;
            int x = getWorldX(i, j), y = getWorldY(i, j);
            getBackground().drawImage(image, x-hexSize/2, y-hexSize/2);
        }
    }
    
    /**
     * Private method to create the image of a hexagon using current state values
     */
    private GreenfootImage getHexImage()
    {
        return getHexImage(hexBackgroundColor, hexFrameColor);
    }
    
    /**
     * Private method to create the image of a hexagon using the given color for the hexagon
     */
    private GreenfootImage getHexImage(Color bg, Color frame)
    {
        int[] xs = { hexSize/4, hexSize*3/4, hexSize, hexSize*3/4, hexSize/4, 0 },
        ys = { 0, 0, hexSize/2, hexSize, hexSize, hexSize/2 };
        if (rotated) { int[] hold = xs; xs = ys; ys = hold; }
        GreenfootImage image = new GreenfootImage(hexSize+1, hexSize+1);
        image.setColor(bg);
        image.fillPolygon(xs, ys, 6);
        image.setColor(frame);
        image.drawPolygon(xs, ys, 6);
        return image;
    }

    /**
     * Returns true if the given hex coordinates are valid for an active hexagon
     *
     * @param col the column number (index of hexagon along the x-axis)
     * @param row the row number (index of hexagon along the y-axis)
     * @return the true/false state of the given hex coordinates being valid for an active hexagon
     */
    public boolean isHexLocation(int col, int row)
    {
        int evenVal = (hexSize%4)/2;
        return !(col < 0 || col >= hexWidth || row < 0 || row >= hexHeight ||
                (rotated && row%2 != evenVal && col == hexWidth-1) ||
                (!rotated && col%2 != evenVal && row == hexHeight-1));
    }
    
    /**
     * Returns the x-coordinate of the center of the hexagon located at the given hex coordinates;
     * the method fails if the given hex coordinates are not valid for an active hexagon
     *
     * @param col the column number (index of hexagon along the x-axis)
     * @param row the row number (index of hexagon along the y-axis)
     * @return the world x-coordinate value for the center of the hexagon at the given hex coordinates
     */
    public int getWorldX(int col, int row)
    {
        if (!isHexLocation(col, row))
        {
            String text = "Index out of range (given:  row: "+row+"   col: "+col+"\n\n";
            text += "The arguments for column and row numbers must correspond to an active hexagon";
            text += "\nat HexWorld.getWorldX\n\n\n";
            // System.out.println(text);
            Greenfoot.stop();
            return -1;
        }
        if (!rotated) return hexSize+col*hexSize*3/4+hexSize/2;
        return hexSize+col*hexSize+((row+(hexSize%4)/2)%2)*hexSize/2+hexSize/2;
    }

    /**
     * Returns the y-coordinate of the center of the hexagon located at the given hex coordinates;
     * the method fails if the given hex coordinates are not valid for an active hexagon
     *
     * @param col the column number (index of hexagon along the x-axis)
     * @param row the row number (index of hexagon along the y-axis)
     * @return the world y-coordinate value for the center of the hexagon at the given hex coordinates
     */
    public int getWorldY(int col, int row)
    {
        if (!isHexLocation(col, row))
        {
            String text = "Index out of range (given:  row: "+row+"   col: "+col+"\n\n";
            text += "The arguments for column and row numbers must correspond to an active hexagon";
            text += "\nat HexWorld.getWorldY\n\n\n";
            // System.out.println(text);
            Greenfoot.stop();
            return -1;
        }
        if (rotated) return hexSize+row*hexSize*3/4+hexSize/2;
        return hexSize+row*hexSize+((col+(hexSize%4)/2)%2)*hexSize/2+hexSize/2;
    }        

    /**
     * Adds an actor into the world at the center of the hexagon located at the given hex coordinates;
     * the method fails if the given hex coordinates are not valid for an active hexagon
     * 
     * @param actor the actor to be added into the world
     * @param col the column number (index of hexagon along the x-axis)
     * @param row the row number (index of hexagon along the y-axis)
     */
    public void addHexActor(Actor actor, int col, int row)
    {
        if (!isHexLocation(col, row))
        {
            String text = "Index out of range (given:  row: "+row+"   col: "+col+"\n\n";
            text += "The arguments for column and row numbers must correspond to an active hexagon";
            text += "\nat HexWorld.addHexActor;  type: "+actor.getClass().getName()+"\n\n\n";
            // System.out.println(text);
            Greenfoot.stop();
            return;
        }
        addObject(actor, getWorldX(col, row), getWorldY(col, row));
    }
    
    /**
     * Returns the current color for the border around the hexagonal area
     *
     * @return the inactive fill color for the region outside the hexagonal area
     */
    public Color getInactiveFillColor()
    {
        return inactiveFillColor;
    }

    /**
     * Use this method to set the color of the outer border
     *
     * @param color the color to use for the outer border
     */
    public void setInactiveFillColor(Color color)
    {
        inactiveFillColor = color;
        paintBackground();
    }
    
    /**
     * Returns the current base background color of the hexagonal area
     *
     * @return the background color of the hexagons
     */
    public Color getHexBackgroundColor()
    {
        return hexBackgroundColor;
    }

    /**
     * Use this method to set the background color of the hexagons
     *
     * @param color the color to use for the background of the hexagons
     */
    public void setHexBackgroundColor(Color color)
    {
        hexBackgroundColor = color;
        paintBackground();
    }
    
    /**
     * Returns the current base frame color of the hexagonal area
     *
     * @return the frame color of the hexagons
     */
    public Color getHexFrameColor()
    {
        return hexFrameColor;
    }

    /**
     * Use this method to set the frame color of the hexagons
     *
     * @param color the color to use for the frames of the hexagons
     */
    public void setHexFrameColor(Color color)
    {
        hexFrameColor = color;
        paintBackground();
    }
    
    /**
     * Sets the background and frame colors of the hexagon at the given hex location to the inactive fill color
     *
     * @param col the column number (index of hex along the x-axis)
     * @param row the row number (index of hex along the y-axis)
     */
    public void setHexInactive(int col, int row)
    {
        setHexColors(inactiveFillColor, hexFrameColor, col, row);
    }
    
    /**
     * Use this method to alter the colors of the hexagon at the given hex coordinates
     * 
     * @param bg the new background color for the hexagon
     * @param frame the new frame color for the hexagon
     * @param col the column number (index of hex along the x-axis)
     * @param row the row number (index of hex along the y-axis)
     */
    public void setHexColors(Color bg, Color frame, int col, int row)
    {
        if (!isHexLocation(col, row))
        {
            String text = "Index out of range (given:  row: "+row+"   col: "+col+"\n\n";
            text += "The arguments for column and row numbers must correspond to an active hexagon";
            text += "\nat HexWorld.setHexColors\n\n\n";
            // System.out.println(text);
            Greenfoot.stop();
            return;
        }
        getBackground().drawImage(getHexImage(bg, frame), getWorldX(col, row)-hexSize/2, getWorldY(col, row)-hexSize/2);
    }
            
    /**
     * Returns the number of columns of hexagons in the world
     * 
     * @return the world hex width (like 'getWidth' for a grid world)
     */
    public int getHexWidth()
    {
        return hexWidth;
    }
    
    /**
     * Returns the number of rows of hexagons in the world
     * 
     * @return the world hex height (like 'getHeight' for a grid world)
     */
    public int getHexHeight()
    {
        return hexHeight;
    }
    
    /**
     * Returns the size (length or width, same size) of the hexagons in the world
     * 
     * @return the size of the hexagons in the world (like 'cellsize' for a grid world)
     */
    public int getHexSize()
    {
        return hexSize;
    }
    
    /**
     * Returns the hex x-coordinate of the location given in world coordinates
     *
     * @param worldX x-coordinate of the location in the world
     * @param worldY y-coordinate of the location in the world
     * @return the hex x-coordinate of the hypothetical hexagon in which the point in the world given in world coordinates resides
     */
    public int getHexX(int worldX, int worldY)
    {
        if (!rotated) return (worldX-hexSize*9/8)/(hexSize*3/4);
        int y = (worldY-hexSize*9/8)/(hexSize*3/4);
        return (worldX-hexSize-(hexSize/2)*((y%2+(hexSize/2)%2)%2))/hexSize;
    }
    
    /**
     * Returns the hex y-coordinate of the location given in world coordinates
     *
     * @param worldX x-coordinate of the location in the world
     * @param worldY y-coordinate of the location in the world
     * @return the hex y-coordinate of the hypothetical hexagon in which the point in the world given in world coordinates resides
     */
    public int getHexY(int worldX, int worldY)
    {
        if (rotated) return (worldY-hexSize*9/8)/(hexSize*3/4);
        int x = (worldX-hexSize*9/8)/(hexSize*3/4);
        return (worldY-hexSize-(hexSize/2)*((x%2+(hexSize/2)%2)%2))/hexSize;
    }
}